from enum import Enum
from enum import IntEnum


class CronType(Enum):
    """The type of event in cron."""

    off = 0


class PowerMode(IntEnum):
    """Power mode of the light."""

    LAST = 0
    NORMAL = 1
    RGB = 2
    HSV = 3
    COLOR_FLOW = 4
    MOONLIGHT = 5


class BulbType(Enum):
    """
    The bulb's type.

    This is either `White` (for monochrome bulbs), `Color` (for color bulbs), `WhiteTemp` (for white bulbs with
    configurable color temperature), `WhiteTempMood` for white bulbs with mood lighting (like the JIAOYUE 650 LED
    ceiling light), or `Unknown` if the properties have not been fetched yet.
    """

    Unknown = -1
    White = 0
    Color = 1
    WhiteTemp = 2
    WhiteTempMood = 3


class LightType(IntEnum):
    """Type of light to control."""

    Main = 0
    Ambient = 1


class SceneClass(IntEnum):
    """
    The scene class to use.

    The scene class (as named in Yeelight docs) specifies how the `Bulb.set_scene` method should act.

    | `COLOR` changes the light to the specified RGB color and brightness.
    | `HSV` changes the light to the specified HSV color and brightness.
    | `CT` changes the light to the specified color temperature.
    | `CF` starts a color flow.
    | `AUTO_DELAY_OFF` turns the light on and sets a timer to turn it back off after the given number of minutes.
    """

    COLOR = 0
    HSV = 1
    CT = 2
    CF = 3
    AUTO_DELAY_OFF = 4



import colorsys
import logging
from enum import Enum
from itertools import chain

from .utils import _clamp

_LOGGER = logging.getLogger(__name__)


class Action(Enum):
    """
    The Flow action enumeration.

    Use this as the ``action`` parameter in a flow, to specify what should
    happen after the flow ends.
    """

    recover = 0
    stay = 1
    off = 2


class Flow(object):
    actions = Action

    def __init__(self, count=0, action=Action.recover, transitions=None):
        """
        A complete flow, consisting of one or multiple transitions.

        Example:
        >>> transitions = [RGBTransition(255, 0, 0), SleepTransition(400)]
        >>> Flow(3, Flow.actions.recover, transitions)

        :param int count: The number of times to run this flow (0 to run
                          forever).
        :param action action: The action to take after the flow stops. Can be
                              ``Flow.actions.recover`` to go back to the state
                              before the flow, ``Flow.actions.stay`` to stay at
                              the last state, and ``Flow.actions.off`` to turn
                              off.
        :param list transitions: A list of :py:class:`FlowTransition
                                 <yeelight.FlowTransition>` instances that
                                 describe the flow transitions to perform.
        """
        if transitions is None:
            transitions = []
        self.count = count
        self.action = action
        self.transitions = transitions

        # Note, main depends on us, so we cannot import BulbException here.
        if len(self.transitions) > 9:
            _LOGGER.warning(
                "The bulb seems to support up to 9 transitions. Your %s might fail."
                % len(self.transitions)
            )

    @property
    def expression(self):
        """
        Return a YeeLight-compatible expression that implements this flow.

        :rtype: list
        """
        expr = chain.from_iterable(
            transition.as_list() for transition in self.transitions
        )
        expr = ", ".join(str(value) for value in expr)
        return expr

    @property
    def as_start_flow_params(self):
        """
        Return a YeeLight start_cf compatible params.

        :rtype: list
        """
        return self.count * len(self.transitions), self.action.value, self.expression


class FlowTransition(object):
    """A single transition in the flow."""

    def as_list(self):
        """
        Return a YeeLight-compatible expression that implements this transition.

        :rtype: list
        """
        brightness = min(int(self.brightness), 100)
        # Duration must be at least 50, otherwise there's an error.
        return [max(50, self.duration), self._mode, self._value, brightness]


class RGBTransition(FlowTransition):
    def __init__(self, red, green, blue, duration=300, brightness=100):
        """
        An RGB transition.

        :param int red: The value of red (0-255).
        :param int green: The value of green (0-255).
        :param int blue: The value of blue (0-255).
        :param int duration: The duration of the effect, in milliseconds. The
                             minimum is 50.
        :param int brightness: The brightness value to transition to (1-100).
        """
        self.red = red
        self.green = green
        self.blue = blue

        # The mode value the YeeLight protocol mandates.
        self._mode = 1

        self.duration = duration
        self.brightness = brightness

    @property
    def _value(self):
        """The YeeLight-compatible value for this transition."""
        red = _clamp(self.red, 0, 255)
        green = _clamp(self.green, 0, 255)
        blue = _clamp(self.blue, 0, 255)
        return red * 65536 + green * 256 + blue

    def __repr__(self):
        return "<%s(%s,%s,%s) duration %s, brightness %s>" % (
            self.__class__.__name__,
            self.red,
            self.green,
            self.blue,
            self.duration,
            self.brightness,
        )


class HSVTransition(FlowTransition):
    def __init__(self, hue, saturation, duration=300, brightness=100):
        """
        An HSV transition.

        :param int hue: The color hue to transition to (0-359).
        :param int saturation: The color saturation to transition to (0-100).
        :param int duration: The duration of the effect, in milliseconds. The
                             minimum is 50.
        :param int brightness: The brightness value to transition to (1-100).
        """
        self.hue = hue
        self.saturation = saturation

        # The mode value the YeeLight protocol mandates.
        self._mode = 1

        self.duration = duration
        self.brightness = brightness

    @property
    def _value(self):
        """The YeeLight-compatible value for this transition."""
        hue = _clamp(self.hue, 0, 359) / 359.0
        saturation = max(0, min(100, self.saturation)) / 100.0

        red, green, blue = [
            int(round(col * 255)) for col in colorsys.hsv_to_rgb(hue, saturation, 1)
        ]
        return red * 65536 + green * 256 + blue

    def __repr__(self):
        return "<%s(%s,%s) duration %s, brightness %s>" % (
            self.__class__.__name__,
            self.hue,
            self.saturation,
            self.duration,
            self.brightness,
        )


class TemperatureTransition(FlowTransition):
    def __init__(self, degrees, duration=300, brightness=100):
        """
        A Color Temperature transition.

        :param int degrees: The degrees to set the color temperature to
                            (1700-6500).
        :param int duration: The duration of the effect, in milliseconds. The
                             minimum is 50.
        :param int brightness: The brightness value to transition to (1-100).
        """
        self.degrees = degrees

        # The mode value the YeeLight protocol mandates.
        self._mode = 2

        self.duration = duration
        self.brightness = _clamp(brightness, 1, 100)

    @property
    def _value(self):
        """The YeeLight-compatible value for this transition."""
        return max(1700, min(6500, self.degrees))

    def __repr__(self):
        return "<%s(%sK) duration %s, brightness %s>" % (
            self.__class__.__name__,
            self.degrees,
            self.duration,
            self.brightness,
        )


class SleepTransition(FlowTransition):
    def __init__(self, duration=300):
        """
        A Sleep transition.

        :param int duration: The duration of the effect, in milliseconds. The
                             minimum is 50.
        """
        # The mode value the YeeLight protocol mandates.
        self._mode = 7

        # Ignored by YeeLight.
        self._value = 1
        self.brightness = 2

        self.duration = duration

    def __repr__(self):
        return "<%s: duration %s>" % (self.__class__.__name__, self.duration)



# encoding: utf8
import asyncio
import contextlib
import json
import logging
import socket
from typing import Dict
from typing import Optional

from .enums import LightType
from .main import _command_to_send_command
from .main import Bulb
from .main import BulbException
from .main import DEFAULT_PROPS

_LOGGER = logging.getLogger(__name__)

TIMEOUT = 15
PING_INTERVAL = 60

KEY_CONNECTED = "connected"

RECONNECT_ERRORS = ("client quota exceeded", "invalid command")
BACKOFF_ERRORS = ("client quota exceeded",)


def _async_command(f):
    """
    A decorator that wraps a function and enables effects.

    This decorator can only be used with async functions
    because it needs to call an awaitable.
    """

    async def wrapper(*args, **kw):
        """A decorator that wraps a function and enables effects."""
        self = args[0]
        cmd = await self.async_send_command(
            *_command_to_send_command(
                self,
                *(await f(*args, **kw)),
                kw.get("effect", self.effect),
                kw.get("duration", self.duration),
                kw.get("power_mode", self.power_mode),
            )
        )
        result = cmd.get("result", [])
        if result:
            return result[0]

    return wrapper


class AsyncBulb(Bulb):
    """Asyncio support for Bulb."""

    def __init__(self, *args, **kwargs) -> None:
        # Asyncio
        super().__init__(*args, **kwargs)
        self._async_callback = None
        self._async_pending_commands: Dict[int, asyncio.Future] = {}
        self._async_listen_task = None
        self._async_reconnect_task = None
        self._async_writer = None
        self._async_reader = None
        self._async_cmd_id = 0
        self._async_command_lock = asyncio.Lock()
        self._socket_backoff = False
        self._listen_event: Optional[asyncio.Event] = None
        self._music_mode_params = None
        # Prevent sending disconnects if disconnecting for music mode connection
        self._music_expect_disconnect = False
        self._music_mode_lock = asyncio.Lock()

    async def async_send_command(self, method, params):
        """Send a command to the bulb and wait for the result."""
        # Prevent changes to music mode while sending a command
        # unless we are currently activating music mode
        # this prevents sending a command while we are trying to connect
        # and causing a hang while waiting for the response
        if self._async_command_lock.locked() and method == "set_music":
            command_lock = contextlib.AsyncExitStack()
        else:
            command_lock = self._async_command_lock
        async with command_lock:
            if self._music_mode_state:
                await self._async_send_command(method, params, create_future=False)
                # We can't check if it worked, so we just assume it did
                if self._async_callback:
                    self._async_callback({"result": ["ok"]})
                return {"result": ["ok"]}
            future = await self._async_send_command(method, params)
            response = await asyncio.wait_for(future, TIMEOUT)

        if "error" in response:
            raise BulbException(response["error"])

        return response

    async def _async_send_command(self, method, params, create_future=True):
        """Send the command."""
        self._async_cmd_id += 1
        request_id = self._async_cmd_id

        if create_future:
            future = asyncio.Future()
            self._async_pending_commands[request_id] = future

            def clean_up(future):
                if future.cancelled():
                    self._async_pending_commands.pop(request_id, None)

            future.add_done_callback(clean_up)

        command = {"id": request_id, "method": method, "params": params}
        request = (json.dumps(command, separators=(",", ":")) + "\r\n").encode("utf8")
        _LOGGER.debug("%s: > %s", self, request)
        if not self._async_writer:
            raise BulbException("The write socket is closed")
        self._async_writer.write(request)
        _LOGGER.debug("%s: Finished _async_send_command", self)
        return future if create_future else request_id

    async def _async_run_listen(self):
        """Backend for async_listen."""
        _LOGGER.debug("%s: Starting listen task", self)
        while self._is_listening:
            try:
                _LOGGER.debug("%s: Starting connection loop", self)
                await self._async_connection_loop()
            finally:
                _LOGGER.debug("%s: Listen task finalizing", self)
                await self._async_close_reader_writer()
                # if its not related to either
                # 1. Music mode turning on/off
                # 2. light being turned off
                if (
                    self._async_callback
                    and not self._music_expect_disconnect
                    and not self._music_mode_lock.locked()
                ):
                    _LOGGER.debug("%s: Sending disconnect callback", self)
                    self._async_callback({KEY_CONNECTED: False})
                if self._is_listening:
                    # Don't backoff if we expect to be disconnected
                    # e.g. if the light is turned off
                    if not self._music_expect_disconnect:
                        await self._async_backoff()
                    await self._async_reconnect_loop()
                self._music_expect_disconnect = False

    async def _async_reconnect_loop(self):
        _LOGGER.debug("%s: Starting reconnect", self)
        while self._is_listening:
            try:
                reader, writer = await asyncio.wait_for(
                    asyncio.open_connection(self._ip, self._port), TIMEOUT
                )
            except (asyncio.TimeoutError, socket.error) as ex:
                _LOGGER.debug(
                    "%s: Reconnected failed with %s, backing off",
                    self,
                    str(ex) or type(ex),
                )
                await asyncio.sleep(TIMEOUT)
            else:
                _LOGGER.debug("%s: Reconnected successfully", self)
                self._async_connected(writer, reader)
                if self._async_callback:
                    self._async_callback({KEY_CONNECTED: True})
                # If connection drops without the light being turned off
                # no longer in music mode if we were in it
                self._music_mode_state = False
                if self._music_mode:
                    # Need to run this separately as starting music mode cancels this task
                    # and starting music mode will require a connection to the bulb
                    asyncio.ensure_future(self.async_start_music(reconnect=True))
                return

        _LOGGER.debug("%s: Reconnect loop stopped", self)

    async def _async_backoff(self):
        """Back off only if we had a previous failure without a success."""
        if self._socket_backoff:
            _LOGGER.debug("%s: Backing off %s seconds", self, TIMEOUT)
            await asyncio.sleep(TIMEOUT)
        self._socket_backoff = True

    async def _async_connection_loop(self) -> None:
        timeouts = 0
        ping_id = -1
        assert self._async_reader is not None
        while self._is_listening:
            try:
                _LOGGER.debug(
                    "%s: Waiting for line, music_mode_state: %s",
                    self,
                    self._music_mode_state,
                )
                if self._listen_event:
                    self._listen_event.set()
                if self._music_mode_state:
                    # Force clear backoff if we are in music mode
                    # We will not receive any messages
                    # so there is no opportunity to back off
                    self._socket_backoff = False
                line = await asyncio.wait_for(
                    self._async_reader.readline(), PING_INTERVAL + TIMEOUT
                )
            except asyncio.TimeoutError:
                # Since we can't get a response from the light in music mode
                # ping the light to keep the connection alive
                if self._music_mode_state:
                    _LOGGER.debug(
                        "%s: Pinging bulb in music mode after %s",
                        self,
                        PING_INTERVAL + TIMEOUT,
                    )
                    if self._last_properties["power"] == "on":
                        await self.async_turn_on()
                    else:
                        # This should rarely happen as music mode is by default turned off
                        # when the light is turned off
                        await self.async_turn_off(is_ping=True)
                    continue
                timeouts += 1
                if timeouts == 2:
                    _LOGGER.debug("%s: Timeout waiting for line", self)
                    return
                _LOGGER.debug(
                    "%s: No data in %s seconds, pinging bulb to make sure its still connected",
                    self,
                    PING_INTERVAL + TIMEOUT,
                )
                ping_id = await self._async_send_command(
                    "get_prop", ["power"], create_future=False
                )
                continue
            except socket.error as ex:
                _LOGGER.debug("%s: Socket error: %s", self, ex)
                return
            except ValueError as ex:
                _LOGGER.debug("%s: Overran buffer: %s", self, ex)
                return
            except BulbException as ex:
                _LOGGER.warning(
                    "%s: Socket unexpectedly closed out from under us: %s", self, ex
                )
                return

            if line and b"\n" not in line:
                _LOGGER.debug("%s: Partial read from bulb: %s", self, line)
                return
            elif line:
                self._socket_backoff = False
                _LOGGER.debug("%s: Success got line: %s", self, line)
                timeouts = 0
            else:
                _LOGGER.debug(
                    "%s: Bulb closed the connection, music_expect_disconnect: %s",
                    self,
                    self._music_expect_disconnect,
                )
                return

            try:
                decoded_line = json.loads(line.decode("utf8").rstrip())
            except ValueError:
                _LOGGER.error("%s: Invalid data: %s", self, line)
                continue

            if "id" in decoded_line:
                future = self._async_pending_commands.pop(decoded_line["id"], None)
                if future:
                    future.set_result(decoded_line)
                elif decoded_line["id"] == ping_id:
                    _LOGGER.debug("%s: Ping result received: %s", self, decoded_line)
                    data = {"power": decoded_line["result"][0]}
                    self._set_last_properties(data, update=True)
                    data[KEY_CONNECTED] = True
                    self._async_callback(data)
                    continue

            if (
                "error" in decoded_line
                and decoded_line["error"].get("message") in RECONNECT_ERRORS
            ):
                message = decoded_line["error"]["message"]
                _LOGGER.debug(
                    "%s: %s, dropping connection and reconnecting", self, message
                )
                if message in BACKOFF_ERRORS:
                    # Force backoff since reconnect will not clear the quota right away
                    self._socket_backoff = True
                return

            if decoded_line.get("method") != "props":
                _LOGGER.debug("%s: props not in line: %s", self, line)
                continue

            # Update notification received
            _LOGGER.debug("%s: New props received: %s", self, decoded_line)
            self._set_last_properties(decoded_line["params"], update=True)
            data = decoded_line["params"]
            data.update({KEY_CONNECTED: True})
            try:
                self._async_callback(data)
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception("Error while processing external callback")

    def _async_connected(self, writer, reader):
        """Called when we are successfully connected to the bulb."""
        self._async_cmd_id = 0
        self._async_writer = writer
        self._async_reader = reader

    async def async_listen(self, callback):
        """
        Listen to state update notifications.

        This function is blocking until a socket error occurred or being stopped by
        ``stop_listening``. It should be run in an ``asyncio`` task.

        The callback function should take one parameter, containing the new/updates
        properties. It will be called when ``last_properties`` is updated.

        Reconnection happens automaticlly if the socket is closed until
        async_stop_listening is called

        :param callable callback: A callback function to receive state update notification.
        """
        self._async_callback = callback
        try:
            reader, writer = await asyncio.wait_for(
                asyncio.open_connection(self._ip, self._port), TIMEOUT
            )
        except asyncio.TimeoutError as ex:
            raise BulbException(
                f"Timed out trying to the the bulb at {self._ip}:{self._port}."
            ) from ex
        except socket.error as ex:
            raise BulbException(
                f"Failed to read from the socket at {self._ip}:{self._port}: {ex}."
            ) from ex

        self._is_listening = True
        self._async_connected(writer, reader)
        self._async_listen_task = asyncio.ensure_future(self._async_run_listen())
        self._async_callback({KEY_CONNECTED: True})

    def _async_stop_listen_task(self):
        if self._async_listen_task:
            self._async_listen_task.cancel()
            self._async_listen_task = None

    async def _async_close_reader_writer(self):
        self._async_pending_commands = {}
        if self._async_writer:
            # Need to ignore socket errors if it was dropped
            with contextlib.suppress(socket.error):
                self._async_writer.close()
                await asyncio.wait_for(
                    self._async_writer.wait_closed(), timeout=TIMEOUT
                )
            self._async_writer = None
        self._async_reader = None

    async def async_stop_listening(self, remove_callback=True):
        """Stop listening to notifications."""
        self._is_listening = False
        self._async_stop_listen_task()
        await self._async_close_reader_writer()
        if remove_callback:
            self._async_callback = None

    async def async_get_properties(
        self, requested_properties=DEFAULT_PROPS, when_on=False
    ):
        """
        Retrieve and return the properties of the bulb.

        This method also updates ``last_properties`` when it is called.

        The ``current_brightness`` property is calculated by the library (i.e. not returned
        by the bulb), and indicates the current brightness of the lamp, aware of night light
        mode. It is 0 if the lamp is off, and None if it is unknown.

        :param list requested_properties: The list of properties to request from the bulb.
                                          By default, this does not include ``flow_params``.
        :param bool when_on: Only refresh if on.

        :returns: A dictionary of param: value items.
        :rtype: dict
        """
        # when_on_music is only called when reconnecting to music mode
        # ignores music mode flag and only refreshes if on
        if (
            when_on and self._last_properties["power"] != "on"
        ) or self._music_mode_state:
            return self._last_properties

        response = await self.async_send_command("get_prop", requested_properties)
        # do a second check here because music mode could have been enabled
        # while the command was queued this would cause an incorrect response
        # {'power': 'ok', 'current_brightness': None}
        if self._music_mode_state:
            return self._last_properties
        if response is not None and "result" in response:
            properties = response["result"]
            properties = [x if x else None for x in properties]
            new_values = dict(zip(requested_properties, properties))

            # this was a music mode response, so ignore this update
            if new_values["power"] == "ok":
                return self._last_properties

            self._set_last_properties(new_values, update=False)

        return self._last_properties

    async def async_ensure_on(self):
        """Turn the bulb on if it is off."""
        if self._music_mode_state is True or self.auto_on is False:
            return

        await self.async_get_properties()

        if self._last_properties["power"] != "on":
            await self.async_turn_on()

    @_async_command
    async def async_set_color_temp(self, degrees, light_type=LightType.Main, **kwargs):
        """
        Set the bulb's color temperature.

        :param int degrees: The degrees to set the color temperature to (min/max are
                            specified by the model's capabilities, or 1700-6500).
        :param yeelight.LightType light_type: Light type to control.
        """
        await self.async_ensure_on()
        return self._set_color_temp(degrees, light_type=light_type, **kwargs)

    @_async_command
    async def async_set_rgb(
        self, red, green, blue, light_type=LightType.Main, **kwargs
    ):
        """
        Set the bulb's RGB value.

        :param int red:   The red value to set (0-255).
        :param int green: The green value to set (0-255).
        :param int blue:  The blue value to set (0-255).
        :param yeelight.LightType light_type:
                          Light type to control.
        """
        await self.async_ensure_on()
        return self._set_rgb(red, green, blue, light_type=light_type, **kwargs)

    @_async_command
    async def async_set_adjust(self, action, prop, **kwargs):
        """
        Adjust a parameter.

        I don't know what this is good for. I don't know how to use it, or why.
        I'm just including it here for completeness, and because it was easy,
        but it won't get any particular love.

        :param str action: The direction of adjustment. Can be "increase",
                           "decrease" or "circle".
        :param str prop:   The property to adjust. Can be "bright" for
                           brightness, "ct" for color temperature and "color"
                           for color. The only action for "color" can be
                           "circle". Why? Who knows.
        """
        return self._set_adjust(action, prop, **kwargs)

    @_async_command
    async def async_set_hsv(
        self, hue, saturation, value=None, light_type=LightType.Main, **kwargs
    ):
        """
        Set the bulb's HSV value.

        :param int hue:        The hue to set (0-359).
        :param int saturation: The saturation to set (0-100).
        :param int value:      The value to set (0-100). If omitted, the bulb's
                               brightness will remain the same as before the
                               change.
        :param yeelight.LightType light_type: Light type to control.
        """
        await self.async_ensure_on()
        return self._set_hsv(hue, saturation, value, light_type, **kwargs)

    @_async_command
    async def async_set_brightness(
        self, brightness, light_type=LightType.Main, **kwargs
    ):
        """
        Set the bulb's brightness.

        :param int brightness: The brightness value to set (1-100).
        :param yeelight.LightType light_type: Light type to control.
        """
        await self.async_ensure_on()
        return self._set_brightness(brightness, light_type=light_type, **kwargs)

    @_async_command
    async def async_turn_on(self, light_type=LightType.Main, **kwargs):
        """
        Turn the bulb on.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._turn_on(light_type=light_type, **kwargs)

    @_async_command
    async def async_turn_off(
        self, light_type=LightType.Main, is_ping: bool = False, **kwargs
    ):
        """
        Turn the bulb off.

        :param yeelight.LightType light_type: Light type to control.
        :param bool is_ping: If true, don't disable music mode.
        """
        # Turning off implicitly disables music mode.
        # Prevent sending disconnects
        if self._music_mode and not is_ping:
            self._music_expect_disconnect = True
            self._music_mode = False
        return self._turn_off(light_type=light_type, **kwargs)

    @_async_command
    async def async_toggle(self, light_type=LightType.Main, **kwargs):
        """
        Toggle the bulb on or off.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._toggle(light_type=light_type, **kwargs)

    @_async_command
    async def async_dev_toggle(self, **kwargs):
        """Toggle the main light and the ambient on or off."""
        return self._dev_toggle(**kwargs)

    @_async_command
    async def async_set_default(self, light_type=LightType.Main, **kwargs):
        """
        Set the bulb's current state as the default, which is what the bulb will be set to on power on.

        If you get a "general error" setting this, yet the bulb reports as supporting `set_default` during
        discovery, disable "auto save settings" in the YeeLight app.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._set_default(light_type=light_type, **kwargs)

    @_async_command
    async def async_set_name(self, name, **kwargs):
        """
        Set the bulb's name.

        :param str name: The string you want to set as the bulb's name.
        """
        return self._set_name(name, **kwargs)

    @_async_command
    async def async_start_flow(self, flow, light_type=LightType.Main, **kwargs):
        """
        Start a flow.

        :param yeelight.Flow flow: The Flow instance to start.
        """
        await self.async_ensure_on()
        return self._start_start_flow(flow, light_type=light_type, **kwargs)

    @_async_command
    async def async_stop_flow(self, light_type=LightType.Main, **kwargs):
        """
        Stop a flow.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._stop_flow(light_type=light_type, **kwargs)

    @_async_command
    async def async_set_scene(
        self, scene_class, *args, light_type=LightType.Main, **kwargs
    ):
        """
        Set the light directly to the specified state.

        If the light is off, it will first be turned on.

        :param yeelight.SceneClass scene_class: The YeeLight scene class to use.

        * `COLOR` changes the light to the specified RGB color and brightness.

            Arguments:
            * **red** (*int*)         – The red value to set (0-255).
            * **green** (*int*)       – The green value to set (0-255).
            * **blue** (*int*)        – The blue value to set (0-255).
            * **brightness** (*int*)  – The brightness value to set (1-100).

        * `HSV` changes the light to the specified HSV color and brightness.

            Arguments:
            * **hue** (*int*)         – The hue to set (0-359).
            * **saturation** (*int*)  – The saturation to set (0-100).
            * **brightness** (*int*)  – The brightness value to set (1-100).

        * `CT` changes the light to the specified color temperature.

            Arguments:
            * **degrees** (*int*)     – The degrees to set the color temperature to (min/max are specified by the
            model's capabilities, or 1700-6500).
            * **brightness** (*int*)  – The brightness value to set (1-100).

        * `CF` starts a color flow.

            Arguments:
            * **flow** (`yeelight.Flow`)  – The Flow instance to start.

        * `AUTO_DELAY_OFF` turns the light on to the specified brightness and sets a timer to turn it back off after the
          given number of minutes.

            Arguments:
            * **brightness** (*int*)     – The brightness value to set (1-100).
            * **minutes** (*int*)        – The minutes to wait before automatically turning the light off.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._set_scene(scene_class, *args, light_type=light_type, **kwargs)

    @_async_command
    async def async_cron_add(self, event_type, value, **kwargs):
        """
        Add an event to cron.

        Example::

        >>> bulb.cron_add(CronType.off, 10)

        :param yeelight.CronType event_type: The type of event. Currently,
                                                   only ``CronType.off``.
        """
        return self._cron_add(event_type, value, **kwargs)

    @_async_command
    async def async_cron_get(self, event_type, **kwargs):
        """
        Retrieve an event from cron.

        :param yeelight.CronType event_type: The type of event. Currently,
                                                   only ``CronType.off``.
        """
        return self._cron_get(event_type, **kwargs)

    @_async_command
    async def cron_del(self, event_type, **kwargs):
        """
        Remove an event from cron.

        :param yeelight.CronType event_type: The type of event. Currently,
                                                   only ``CronType.off``.
        """
        return self._cron_del(event_type, **kwargs)

    def __repr__(self):
        return "AsyncBulb<{ip}:{port}, type={type}, model={model}>".format(
            ip=self._ip, port=self._port, type=self.bulb_type, model=self.model
        )

    async def async_set_power_mode(self, mode):
        """
        Set the light power mode.

        If the light is off it will be turned on.

        :param yeelight.PowerMode mode: The mode to switch to.
        """
        return await self.async_turn_on(power_mode=mode)

    async def async_start_music(self, port=0, ip=None, reconnect=False):
        """
        Start music mode.

        Music mode essentially upgrades the existing connection to a reverse one
        (the bulb connects to the library), removing all limits and allowing you
        to send commands without being rate-limited.

        Starting music mode will start a new listening socket, tell the bulb to
        connect to that, and then close the old connection. If the bulb cannot
        connect to the host machine for any reason, bad things will happen (such
        as library freezes).

        :param int port: The port to listen on. If none is specified, a random
                         port will be chosen.

        :param str ip: The IP address of the host this library is running on.
                       Will be discovered automatically if not provided.

        :param bool reconnect: Whether to reconnect to the bulb after disconnecting
        """
        async with self._music_mode_lock:
            if self._music_mode and not reconnect:
                # Music mode is enabled and we're already connected.
                raise AssertionError(
                    "Already in music mode, please stop music mode first."
                )

            if self._music_mode_state:
                _LOGGER.debug("%s: Already in music mode", self)
                return "ok"

            if reconnect:
                _LOGGER.debug("%s: Starting music mode reconnect", self)
                # Music mode is enabled but we're not connected.
                # Attempt to retrieve original parameters passed
                port, ip = self._music_mode_params or (port, ip)
                # Attempt to retrieve latest properties if on
                await self.async_get_properties(when_on=True)
            else:
                _LOGGER.debug("%s: Starting music mode", self)
                self._music_mode = True
                # Force populating the cache in case we are being called directly
                # without ever fetching properties beforehand
                await self.async_get_properties()

            future = asyncio.Future()

            # The bulb doesn't send anything in music mode

            def on_connect(reader, writer):
                server.close()
                future.set_result((reader, writer))

            local_ip = ip if ip else self._socket.getsockname()[0]
            server = await asyncio.start_server(
                on_connect, local_ip, port, reuse_address=True
            )
            port = server.sockets[0].getsockname()[1]
            # hold the lock until we are fully enabled
            async with self._async_command_lock:
                await self.async_send_command("set_music", [1, local_ip, port])
                await self.async_stop_listening(False)
                try:
                    reader, writer = await asyncio.wait_for(
                        asyncio.shield(future), timeout=0.5
                    )
                except asyncio.TimeoutError:
                    # send a disconnected callback if we can't connect quickly
                    # then continue waiting to try to connect
                    _LOGGER.debug("%s: Failed to connect to music mode quickly", self)
                    if self._async_callback:
                        self._async_callback({KEY_CONNECTED: False})
                    try:
                        reader, writer = await asyncio.wait_for(
                            future, timeout=TIMEOUT - 0.5
                        )
                    except asyncio.TimeoutError as ex:
                        # Ensures a full reconnect to the bulb
                        await self.async_stop_music(force=True)
                        raise BulbException(
                            f"Timed out enabling music mode on the bulb at {self._ip}:{self._port}."
                        ) from ex
                # Manually enable listener to watch for disconnects
                self._is_listening = True
                self._async_connected(writer, reader)
                self._music_mode_state = True
                self._listen_event = asyncio.Event()
                self._async_listen_task = asyncio.ensure_future(
                    self._async_run_listen()
                )
                try:
                    await asyncio.wait_for(self._listen_event.wait(), timeout=TIMEOUT)
                except asyncio.TimeoutError:
                    # this shouldn't ever happen
                    _LOGGER.debug("%s: Listener failed to start in music mode", self)
                    await self.async_stop_music(force=True)
                    if self._async_callback:
                        self._async_callback({KEY_CONNECTED: False})
                    raise BulbException(
                        f"Timed out waiting for listener on the bulb at {self._ip}:{self._port}."
                    )
            # We are now connected
            if self._async_callback:
                self._async_callback({KEY_CONNECTED: True})

            self._music_mode_params = (port, ip)

            if reconnect:
                _LOGGER.debug("%s: Music mode reconnected successfully", self)
            else:
                _LOGGER.debug("%s: Music mode started successfully", self)

            return "ok"

    async def async_stop_music(self, force=False, **kwargs):
        """
        Stop music mode.

        Stopping music mode will close the previous connection. Calling
        ``stop_music`` more than once, or while not in music mode, is safe.
        :param bool force: Whether to force stop music mode (during exception when enabling)
        """
        self._music_mode = False

        if not self._music_mode_state and not force:
            _LOGGER.debug(
                "%s: Music mode was not enabled but async_stop_music was called",
                self,
            )
            return

        # flush out music mode socket
        self._music_expect_disconnect = True
        await self.async_stop_listening(False)
        self._music_expect_disconnect = False
        self._music_mode_state = False

        await self.async_listen(self._async_callback)
        return "set_music", [0], kwargs



"""Pre-made transitions, for your strobing pleasure."""
import random

from .flow import HSVTransition
from .flow import RGBTransition
from .flow import SleepTransition
from .flow import TemperatureTransition
from .utils import _clamp


def disco(bpm=120):
    """
    Color changes to the beat.

    :param int bpm: The beats per minute to pulse to.

    :returns: A list of transitions.
    :rtype: list
    """
    duration = int(60000 / bpm)
    transitions = [
        HSVTransition(0, 100, duration=duration, brightness=100),
        HSVTransition(0, 100, duration=duration, brightness=1),
        HSVTransition(90, 100, duration=duration, brightness=100),
        HSVTransition(90, 100, duration=duration, brightness=1),
        HSVTransition(180, 100, duration=duration, brightness=100),
        HSVTransition(180, 100, duration=duration, brightness=1),
        HSVTransition(270, 100, duration=duration, brightness=100),
        HSVTransition(270, 100, duration=duration, brightness=1),
    ]
    return transitions


def temp():
    """
    Slowly-changing color temperature.

    :returns: A list of transitions.
    :rtype: list
    """
    transitions = [
        TemperatureTransition(1700, duration=40000),
        TemperatureTransition(6500, duration=40000),
    ]
    return transitions


def strobe():
    """
    Rapid flashing on and off.

    :returns: A list of transitions.
    :rtype: list
    """
    transitions = [
        HSVTransition(0, 0, duration=50, brightness=100),
        HSVTransition(0, 0, duration=50, brightness=1),
    ]
    return transitions


def pulse(red, green, blue, duration=250, brightness=100):
    """
    Pulse a single color once (mainly to be used for notifications).

    :param int red: The red color component to pulse (0-255).
    :param int green: The green color component to pulse (0-255).
    :param int blue: The blue color component to pulse (0-255).
    :param int duration: The duration to pulse for, in milliseconds.
    :param int brightness: The brightness to pulse at (1-100).

    :returns: A list of transitions.
    :rtype: list
    """
    red = _clamp(red, 0, 255)
    green = _clamp(green, 0, 255)
    blue = _clamp(blue, 0, 255)

    transitions = [
        RGBTransition(red, green, blue, duration=duration, brightness=brightness),
        RGBTransition(red, green, blue, duration=duration, brightness=1),
    ]
    return transitions


def strobe_color(brightness=100):
    """
    Rapid flashing colors.

    :param int brightness: The brightness of the transition.

    :returns: A list of transitions.
    :rtype: list
    """
    transitions = [
        HSVTransition(240, 100, duration=50, brightness=brightness),
        HSVTransition(60, 100, duration=50, brightness=brightness),
        HSVTransition(330, 100, duration=50, brightness=brightness),
        HSVTransition(0, 100, duration=50, brightness=brightness),
        HSVTransition(173, 100, duration=50, brightness=brightness),
        HSVTransition(30, 100, duration=50, brightness=brightness),
    ]
    return transitions


def alarm(duration=250):
    """
    Red alarm; flashing bright red to dark red.

    :param int duration: The duration between hi/lo brightness,in milliseconds.

    :returns: A list of transitions.
    :rtype: list
    """
    transitions = [
        HSVTransition(0, 100, duration=duration, brightness=100),
        HSVTransition(0, 100, duration=duration, brightness=60),
    ]
    return transitions


def police(duration=300, brightness=100):
    """
    Color changes from red to blue, like police lights.

    :param int duration: The duration between red and blue, in milliseconds.
    :param int brightness: The brightness of the transition.

    :returns: A list of transitions.
    :rtype: list
    """
    transitions = [
        RGBTransition(255, 0, 0, duration=duration, brightness=brightness),
        RGBTransition(0, 0, 255, duration=duration, brightness=brightness),
    ]
    return transitions


def police2(duration=250, brightness=100):
    """
    Color flashes red and then blue, like urgent police lights.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.

    :returns: A list of transitions.
    :rtype: list
    """
    transitions = [
        RGBTransition(255, 0, 0, brightness=brightness, duration=duration),
        RGBTransition(255, 0, 0, brightness=1, duration=duration),
        RGBTransition(255, 0, 0, brightness=brightness, duration=duration),
        SleepTransition(duration=duration),
        RGBTransition(0, 0, 255, brightness=brightness, duration=duration),
        RGBTransition(0, 0, 255, brightness=1, duration=duration),
        RGBTransition(0, 0, 255, brightness=brightness, duration=duration),
        SleepTransition(duration=duration),
    ]
    return transitions


def lsd(duration=3000, brightness=100):
    """
    Gradual changes to a pleasing, trippy palette.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.

    :returns: A list of transitions.
    :rtype: list
    """
    hs_values = [(3, 85), (20, 90), (55, 95), (93, 50), (198, 97)]
    return [
        HSVTransition(hue, saturation, duration=duration, brightness=brightness)
        for hue, saturation in hs_values
    ]


def christmas(duration=250, brightness=100, sleep=3000):
    """
    Color changes from red to green, like christmas lights.

    :param int duration: The duration between red and green, in milliseconds.
    :param int brightness: The brightness of the transition.
    :param int sleep: The time to sleep between colors, in milliseconds.

    :returns: A list of transitions.
    :rtype: list
    """
    transitions = [
        HSVTransition(0, 100, duration=duration, brightness=brightness),
        SleepTransition(duration=sleep),
        HSVTransition(120, 100, duration=duration, brightness=brightness),
        SleepTransition(duration=sleep),
    ]
    return transitions


def rgb(duration=250, brightness=100, sleep=3000):
    """
    Color changes from red to green to blue.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.
    :param int sleep: The time to sleep between colors, in milliseconds

    :returns: A list of transitions.
    :rtype: list
    """
    transitions = [
        HSVTransition(0, 100, duration=duration, brightness=brightness),
        SleepTransition(duration=sleep),
        HSVTransition(120, 100, duration=duration, brightness=brightness),
        SleepTransition(duration=sleep),
        HSVTransition(240, 100, duration=duration, brightness=brightness),
        SleepTransition(duration=sleep),
    ]
    return transitions


def random_loop(duration=750, brightness=100, count=9):
    """
    Color changes between `count` randomly chosen colors.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.
    :param int count: The number of random chosen colors in transition.

    :returns: A list of transitions.
    :rtype: list
    """
    count = _clamp(count, 1, 9)
    transitions = [
        HSVTransition(
            random.randint(0, 360), 100, duration=duration, brightness=brightness
        )
        for _ in range(count)
    ]
    return transitions


def slowdown(duration=2000, brightness=100, count=8):
    """
    Change between `count` random chosen colors with increasing transition time.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.
    :param int count: The number of random chosen colors in transition.

    :returns: A list of transitions.
    :rtype: list
    """
    count = _clamp(count, 1, 8)
    transitions = [
        HSVTransition(
            random.randint(0, 360), 100, duration=(duration * x), brightness=brightness
        )
        for x in range(1, count + 1)
    ]
    return transitions



# encoding: utf8
import colorsys
import json
import logging
import socket

from future.utils import raise_from

from .decorator import decorator  # type: ignore
from .enums import BulbType
from .enums import LightType
from .enums import PowerMode
from .enums import SceneClass
from .flow import Flow
from .ssdp_discover import filter_lower_case_keys
from .ssdp_discover import parse_capabilities
from .ssdp_discover import send_discovery_packet
from .utils import _clamp
from .utils import rgb_to_yeelight


try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse  # type: ignore

_LOGGER = logging.getLogger(__name__)


DEFAULT_PROPS = [
    "power",
    "bright",
    "ct",
    "rgb",
    "hue",
    "sat",
    "color_mode",
    "flowing",
    "delayoff",
    "music_on",
    "name",
    "bg_power",
    "bg_flowing",
    "bg_ct",
    "bg_bright",
    "bg_hue",
    "bg_sat",
    "bg_rgb",
    "nl_br",
    "active_mode",
]

_MODEL_SPECS = {
    "bslamp1": {
        "color_temp": {"min": 1700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "bslamp2": {
        "color_temp": {"min": 1700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "bslamp3": {
        "color_temp": {"min": 1700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "ceil26": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "ceila": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "ceilc": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": True,
    },
    "ceiling10": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": True,
    },
    "ceiling13": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "ceiling15": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "ceiling18": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "ceiling19": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": True,
    },
    "ceiling1": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "ceiling20": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": True,
    },
    "ceiling24": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "ceiling2": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "ceiling3": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "ceiling4": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": True,
    },
    "ceiling5": {
        "color_temp": {"min": 2700, "max": 5700},
        "night_light": True,
        "background_light": False,
    },
    "ceiling6": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": True,
        "background_light": False,
    },
    "color1": {
        "color_temp": {"min": 1700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "color2": {
        "color_temp": {"min": 1700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "color4": {
        "color_temp": {"min": 1700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "colorc": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "color": {
        "color_temp": {"min": 1700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "ct_bulb": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "ct2": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "lamp1": {
        "color_temp": {"min": 2700, "max": 5000},
        "night_light": False,
        "background_light": False,
    },
    "lamp3": {
        "color_temp": {"min": 4000, "max": 4000},
        "night_light": False,
        "background_light": False,
        "bulb_type": BulbType.White,
    },
    "lamp4": {
        "color_temp": {"min": 2600, "max": 5000},
        "night_light": False,
        "background_light": False,
    },
    "lamp15": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": False,
        "background_light": True,
        "bulb_type": BulbType.WhiteTempMood,
    },
    "mono1": {
        "color_temp": {"min": 2700, "max": 2700},
        "night_light": False,
        "background_light": False,
    },
    "mono5": {
        "color_temp": {"min": 2700, "max": 2700},
        "night_light": False,
        "background_light": False,
    },
    "mono": {
        "color_temp": {"min": 2700, "max": 2700},
        "night_light": False,
        "background_light": False,
    },
    "monob": {
        "color_temp": {"min": 2700, "max": 2700},
        "night_light": False,
        "background_light": False,
    },
    "strip1": {
        "color_temp": {"min": 1700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "strip2": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "strip4": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
    "strip6": {
        "color_temp": {"min": 2700, "max": 6500},
        "night_light": False,
        "background_light": False,
    },
}


def get_known_models():
    """
    Helper method to return all known yeelight models.

    The models spec dict is private and internal, this function allows consumers to get
    a list of models via a public method.
    """
    return list(_MODEL_SPECS.keys())


def _command_to_send_command(
    self, method, params, kwargs, effect, duration, power_mode
):
    """
    Convert args and kwargs to method and params.

    This provides the underlying fuctionality for the _command
    decorator and aio._async_command decorator. This function
    contains the code that can be shared between the sync
    and async versions.
    """
    light_type = kwargs.get("light_type", LightType.Main)

    # Prepend the control for different bulbs
    if light_type == LightType.Ambient:
        method = "bg_" + method

    if method in [
        "set_ct_abx",
        "set_rgb",
        "set_hsv",
        "set_bright",
        "set_power",
        "toggle",
        "bg_set_ct_abx",
        "bg_set_rgb",
        "bg_set_hsv",
        "bg_set_bright",
        "bg_set_power",
        "bg_toggle",
    ]:
        if self._music_mode or self._music_mode_state:
            # Mapping calls to their properties.
            # Used to keep music mode cache up to date.
            action_property_map = {
                "set_ct_abx": ["ct"],
                "bg_set_ct_abx": ["bg_ct"],
                "set_rgb": ["rgb"],
                "bg_set_rgb": ["bg_rgb"],
                "set_hsv": ["hue", "sat"],
                "bg_set_hsv": ["bg_hue", "bg_sat"],
                "set_bright": ["bright"],
                "bg_set_bright": ["bg_bright"],
                "set_power": ["power"],
                "bg_set_power": ["bg_power"],
            }
            # Handle toggling separately, as it depends on a previous power state.
            if method == "toggle":
                self._last_properties["power"] = (
                    "on" if self._last_properties["power"] == "off" else "off"
                )
            if method == "bg_toggle":
                self._last_properties["bg_power"] = (
                    "on" if self._last_properties["bg_power"] == "off" else "off"
                )
            # dev_toggle toggle both lights depending on the MAIN light power status.
            if method == "dev_toggle":
                new_state = "on" if self._last_properties["power"] == "off" else "off"
                self._last_properties["power"] = new_state
                self._last_properties["bg_power"] = new_state
            elif method in action_property_map:
                set_prop = action_property_map[method]
                update_props = {
                    set_prop[prop]: params[prop] for prop in range(len(set_prop))
                }
                if "rgb" in method:
                    update_props["color_mode"] = 1
                elif "ct" in method:
                    update_props["color_mode"] = 2
                elif "hsv" in method:
                    update_props["color_mode"] = 3
                _LOGGER.debug("Music mode cache update: %s", update_props)
                self._last_properties.update(update_props)
        # Add the effect parameters.
        params += [effect, duration]
        # Add power_mode parameter.
        if (
            method == "set_power"
            and params[0] == "on"
            and power_mode.value != PowerMode.LAST
        ):
            params += [power_mode.value]
        if (
            method == "bg_set_power"
            and params[0] == "on"
            and power_mode.value != PowerMode.LAST
        ):
            params += [power_mode.value]

    return method, params


@decorator
def _command(f, *args, **kw):
    """A decorator that wraps a function and enables effects."""
    self = args[0]
    cmd = self.send_command(
        *_command_to_send_command(
            self,
            *f(*args, **kw),
            kw.get("effect", self.effect),
            kw.get("duration", self.duration),
            kw.get("power_mode", self.power_mode)
        )
    )
    result = cmd.get("result", [])
    if result:
        return result[0]


def discover_bulbs(timeout=2, interface=False):
    """
    Discover all the bulbs in the local network.

    :param int timeout: How many seconds to wait for replies. Discovery will
                        always take exactly this long to run, as it can't know
                        when all the bulbs have finished responding.

    :param string interface: The interface that should be used for multicast packets.
                             Note: it *has* to have a valid IPv4 address. IPv6-only
                             interfaces are not supported (at the moment).
                             The default one will be used if this is not specified.

    :returns: A list of dictionaries, containing the ip, port and capabilities
              of each of the bulbs in the network.
    """
    s = send_discovery_packet(timeout, interface)

    bulbs = []
    bulb_ips = set()
    while True:
        try:
            data, addr = s.recvfrom(65507)
        except socket.timeout:
            break

        capabilities = parse_capabilities(data)
        parsed_url = urlparse(capabilities["Location"])

        bulb_ip = (parsed_url.hostname, parsed_url.port)
        if bulb_ip in bulb_ips:
            continue

        capabilities = filter_lower_case_keys(capabilities)
        bulbs.append(
            {"ip": bulb_ip[0], "port": bulb_ip[1], "capabilities": capabilities}
        )
        bulb_ips.add(bulb_ip)

    return bulbs


class BulbException(Exception):
    """
    A generic yeelight exception.

    This exception is raised when bulb informs about errors, e.g., when trying
    to issue unsupported commands to the bulb.
    """

    pass


class Bulb(object):
    def __init__(
        self,
        ip,
        port=55443,
        effect="smooth",
        duration=300,
        auto_on=False,
        power_mode=PowerMode.LAST,
        model=None,
    ):
        """
        The main controller class of a physical YeeLight bulb.

        :param str ip:       The IP of the bulb.
        :param int port:     The port to connect to on the bulb.
        :param str effect:   The type of effect. Can be "smooth" or "sudden".
        :param int duration: The duration of the effect, in milliseconds. The
                             minimum is 30. This is ignored for sudden effects.
        :param bool auto_on: Whether to call :py:meth:`ensure_on()
                             <yeelight.Bulb.ensure_on>` to turn the bulb on
                             automatically before each operation, if it is off.
                             This renews the properties of the bulb before each
                             message, costing you one extra message per command.
                             Turn this off and do your own checking with
                             :py:meth:`get_properties()
                             <yeelight.Bulb.get_properties()>` or run
                             :py:meth:`ensure_on() <yeelight.Bulb.ensure_on>`
                             yourself if you're worried about rate-limiting.
        :param yeelight.PowerMode power_mode:
                             The mode for the light set when powering on.
        :param str model:    The model name of the yeelight (e.g. "color",
                             "mono", etc). The setting is used to enable model
                             specific features (e.g. a particular color
                             temperature range).

        """
        self._ip = ip
        self._port = port

        self.effect = effect
        self.duration = duration
        self.auto_on = auto_on
        self.power_mode = power_mode
        self._model = model

        self.__cmd_id = 0  # The last command id we used.
        self._last_properties = {}  # The last set of properties we've seen.
        self._capabilities = {}  # Capabilites obtained via SSDP Discovery.
        self._music_mode = False  # Whether we're currently in music mode.
        # When aio is used, this additional variable is used internally
        # because aio tracks the intended music mode/not music mode state
        # as well as the actual bulb connection state. This is required
        # because the sync module does not update or use this variable.
        # In aio, _music_mode is used as the wanted state.
        # _music_mode_state is the current bulb connection type.
        self._music_mode_state = False

        self.__socket = None  # The socket we use to communicate.

        self._notification_socket = None  # The socket to get update notifications
        self._is_listening = False  # Indicate if we are listening

    @property
    def _cmd_id(self):
        """
        Return the next command ID and increment the counter.

        :rtype: int
        """
        self.__cmd_id += 1
        return self.__cmd_id - 1

    @property
    def _socket(self):
        """Return, optionally creating, the communication socket."""
        if self.__socket is None:
            self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.__socket.settimeout(5)
            self.__socket.connect((self._ip, self._port))
        return self.__socket

    def get_capabilities(self, timeout=2):
        """
        Get the bulb's capabilities using the discovery protocol.

        :param int timeout: How many seconds to wait for replies. Discovery will
                            always take exactly this long to run, as it can't know
                            when all the bulbs have finished responding.

        :returns: Dictionary, containing the ip, port and capabilities. For example:
                  {
                  'id': '0x0000000002eb9f61',
                  'model': 'ceiling3',
                  'fw_ver': '43',
                  'support': 'get_prop set_default set_power toggle set_bright set_scene cron_add cron_get cron_del start_cf stop_cf set_ct_abx set_name set_adjust adjust_bright adjust_ct',
                  'power': 'on',
                  'bright': '99',
                  'color_mode': '2',
                  'ct': '3802',
                  'rgb': '0',
                  'hue': '0',
                  'sat': '0',
                  'name': ''
                  }
        """
        s = send_discovery_packet(timeout, ip_address=self._ip)

        try:
            data, addr = s.recvfrom(65507)
        except socket.timeout:
            return None

        capabilities = parse_capabilities(data)
        capabilities = filter_lower_case_keys(capabilities)

        self._capabilities = capabilities
        return capabilities

    def set_capabilities(self, capabilities):
        """Set capabilities from external discovery."""
        self._capabilities = capabilities

    def ensure_on(self):
        """Turn the bulb on if it is off."""
        if self._music_mode is True or self.auto_on is False:
            return

        self.get_properties()

        if self._last_properties["power"] != "on":
            self.turn_on()

    @property
    def last_properties(self):
        """
        The last properties we've seen the bulb have.

        This might potentially be out of date, as there's no background listener
        for the bulb's notifications. To update it, call
        :py:meth:`get_properties <yeelight.Bulb.get_properties()>`.
        """
        return self._last_properties

    @property
    def capabilities(self):
        """
        Capabilities obtained via SSDP Discovery.

        They will be empty, unless updated via:
        :py:meth:`get_capabilities <yeelight.Bulb.get_capabilities()>`.

        :return: Capabilities dict returned by :py:meth:`get_capabilities`.
        """
        return self._capabilities

    @property
    def bulb_type(self):
        """
        The type of bulb we're communicating with.

        Returns a :py:class:`BulbType <yeelight.BulbType>` describing the bulb
        type.

        When trying to access before properties are known, the bulb type is unknown.

        :rtype: yeelight.BulbType
        :return: The bulb's type.
        """
        if not self._last_properties or any(
            name not in self.last_properties for name in ["ct", "rgb"]
        ):
            return BulbType.Unknown

        # Override autodetection if bulb_type is provided in _MODEL_SPECS.
        # We don't use get_model_specs() here to avoid a possible recursion.
        if (
            self.model is not None
            and self.model in _MODEL_SPECS
            and "bulb_type" in _MODEL_SPECS[self.model]
        ):
            return _MODEL_SPECS[self.model]["bulb_type"]

        if "support" in self._capabilities:
            support = set(self._capabilities["support"].split(" "))
            if "set_rgb" in support:
                return BulbType.Color
            if "bg_set_power" in support:
                return BulbType.WhiteTempMood
            if "set_ct_abx" in support:
                return BulbType.WhiteTemp
            return BulbType.White

        if self.last_properties["rgb"] is None and self.last_properties["ct"]:
            if self.last_properties["bg_power"] is not None:
                return BulbType.WhiteTempMood
            else:
                return BulbType.WhiteTemp
        if all(
            name in self.last_properties and self.last_properties[name] is None
            for name in ["ct", "rgb", "hue", "sat"]
        ):
            return BulbType.White
        else:
            return BulbType.Color

    @property
    def model(self):
        """
        Return declared model / model discovered via SSDP Discovery or None.

        :return: Device model
        """
        if self._model:
            return self._model
        elif "model" in self.capabilities:
            return self.capabilities["model"]
        else:
            return None

    @property
    def music_mode(self):
        """
        Return whether the music mode is active.

        :rtype: bool
        :return: True if music mode is on, False otherwise.
        """
        return self._music_mode

    @property
    def music_mode_state(self):
        """
        Return whether the connection is music mode (aio only).

        When using the sync module, this variable is ignored and not updated.

        :rtype: bool
        :return: True if music mode is connected, False otherwise.
        """
        return self._music_mode_state

    def listen(self, callback):
        """
        Listen to state update notifications.

        This function is blocking until a socket error occurred or being stopped by
        ``stop_listening``. It should be run in a ``Thread`` or ``asyncio`` event loop.

        The callback function should take one parameter, containing the new/updates
        properties. It will be called when ``last_properties`` is updated.

        :param callable callback: A callback function to receive state update notification.
        """
        try:
            self._is_listening = True
            self._notification_socket = socket.socket(
                socket.AF_INET, socket.SOCK_STREAM
            )
            self._notification_socket.setblocking(True)
            self._notification_socket.connect((self._ip, self._port))
            while self._notification_socket is not None:
                data = self._notification_socket.recv(16 * 1024)
                for line in data.split(b"\r\n"):
                    if not line:
                        continue
                    try:
                        line = json.loads(line.decode("utf8"))
                    except ValueError:
                        _LOGGER.error("Invalid data: %s", line)
                        continue

                    if line.get("method") == "props":
                        # Update notification received
                        _LOGGER.debug("New props received: %s", line)
                        self._set_last_properties(line["params"], update=True)
                        callback(line["params"])
        except socket.error as ex:
            if not self._is_listening:
                # Socket is manually shutdown by stop_listening
                return
            self._notification_socket.close()
            self._notification_socket = None
            raise_from(BulbException("Failed to read from the socket."), ex)

    def stop_listening(self):
        """Stop listening to notifications."""
        self._is_listening = False
        self._notification_socket.shutdown(socket.SHUT_RDWR)
        self._notification_socket.close()
        self._notification_socket = None

    def _set_last_properties(self, properties, update=True):
        """Update derived properties after an update of the self._last_properties."""
        if update:
            self._last_properties.update(properties)
        else:
            self._last_properties = properties

        if self._last_properties.get("power") == "off":
            cb = "0"
        if self._last_properties.get("bg_power") == "off":
            cb = "0"
        elif self._last_properties.get("active_mode") == "1":
            # Nightlight mode.
            cb = self._last_properties.get("nl_br")
        else:
            cb = self._last_properties.get("bright")
        self._last_properties["current_brightness"] = cb

    def get_properties(
        self,
        requested_properties=DEFAULT_PROPS,
        ssdp_fallback=False,
    ):
        """
        Retrieve and return the properties of the bulb.

        This method also updates ``last_properties`` when it is called.

        The ``current_brightness`` property is calculated by the library (i.e. not returned
        by the bulb), and indicates the current brightness of the lamp, aware of night light
        mode. It is 0 if the lamp is off, and None if it is unknown.

        :param list requested_properties: The list of properties to request from the bulb.
                                          By default, this does not include ``flow_params``.
        :param bool ssdp_fallback: Fallback to SSDP should get_prop fail,
                                    does not work in all network environment, default = False

        :returns: A dictionary of param: value items.
        :rtype: dict
        """
        # When we are in music mode, the bulb does not respond to queries
        # therefore we need to keep the state up-to-date ourselves
        if self._music_mode:
            return self._last_properties

        response = self.send_command("get_prop", requested_properties)
        if response is not None and "result" in response:
            properties = response["result"]
            properties = [x if x else None for x in properties]
            new_values = dict(zip(requested_properties, properties))
        elif ssdp_fallback:
            capabilities = self.get_capabilities()
            new_values = {
                k: capabilities[k] for k in requested_properties if k in capabilities
            }

        self._set_last_properties(new_values, update=False)

        return self._last_properties

    def send_command(self, method, params=None):
        """
        Send a command to the bulb.

        :param str method:  The name of the method to send.
        :param list params: The list of parameters for the method.

        :raises BulbException: When the bulb indicates an error condition.
        :returns: The response from the bulb.
        """
        command = {"id": self._cmd_id, "method": method, "params": params}
        request = (json.dumps(command, separators=(",", ":")) + "\r\n").encode("utf8")
        _LOGGER.debug("%s > %s", self, request)

        try:
            self._socket.send(request)
            # This is a workaround for some firmware versions, it seems to help with the
            # "Bulb closed the connection" issue. More discussion can be found in issue #61.
            self._socket.send(b" ")
        except socket.error as ex:
            # Some error occurred, remove this socket in hopes that we can later
            # create a new one.
            self.__socket.close()
            self.__socket = None
            raise_from(
                BulbException("A socket error occurred when sending the command."), ex
            )

        if self._music_mode:
            # We're in music mode, nothing else will happen.
            return {"result": ["ok"]}

        # The bulb will send us updates on its state in addition to responses,
        # so we want to make sure that we read until we see an actual response.
        response = None
        while response is None:
            try:
                data = self._socket.recv(16 * 1024)
            except socket.error:
                # An error occured, let's close and abort...
                self.__socket.close()
                self.__socket = None
                response = {"error": "Bulb closed the connection."}
                break

            for line in data.split(b"\r\n"):
                if not line:
                    continue

                try:
                    line = json.loads(line.decode("utf8"))
                    _LOGGER.debug("%s < %s", self, line)
                except ValueError:
                    line = {"result": ["invalid command"]}

                if line.get("method") != "props":
                    # This is probably the response we want.
                    response = line
                else:
                    self._last_properties.update(line["params"])

        if (
            method == "set_music"
            and params == [0]
            and "error" in response
            and response["error"]["code"] == -5000
        ):
            # The bulb seems to throw an error for no reason when stopping music mode,
            # it doesn't affect operation and we can't do anything about it, so we might
            # as well swallow it.
            return {"id": 1, "result": ["ok"]}

        if "error" in response:
            raise BulbException(response["error"])

        return response

    @_command
    def set_color_temp(self, degrees, light_type=LightType.Main, **kwargs):
        """
        Set the bulb's color temperature.

        :param int degrees: The degrees to set the color temperature to (min/max are
                            specified by the model's capabilities, or 1700-6500).
        :param yeelight.LightType light_type: Light type to control.
        """
        self.ensure_on()
        return self._set_color_temp(degrees, light_type=light_type, **kwargs)

    def _set_color_temp(self, degrees, light_type=LightType.Main, **kwargs):
        return (
            "set_ct_abx",
            [self._clamp_color_temp(degrees)],
            dict(kwargs, light_type=light_type),
        )

    @_command
    def set_rgb(self, red, green, blue, light_type=LightType.Main, **kwargs):
        """
        Set the bulb's RGB value.

        :param int red:   The red value to set (0-255).
        :param int green: The green value to set (0-255).
        :param int blue:  The blue value to set (0-255).
        :param yeelight.LightType light_type:
                          Light type to control.
        """
        self.ensure_on()
        return self._set_rgb(red, green, blue, light_type=light_type, **kwargs)

    def _set_rgb(self, red, green, blue, light_type=LightType.Main, **kwargs):
        return (
            "set_rgb",
            [rgb_to_yeelight(red, green, blue)],
            dict(kwargs, light_type=light_type),
        )

    @_command
    def set_adjust(self, action, prop, **kwargs):
        """
        Adjust a parameter.

        I don't know what this is good for. I don't know how to use it, or why.
        I'm just including it here for completeness, and because it was easy,
        but it won't get any particular love.

        :param str action: The direction of adjustment. Can be "increase",
                           "decrease" or "circle".
        :param str prop:   The property to adjust. Can be "bright" for
                           brightness, "ct" for color temperature and "color"
                           for color. The only action for "color" can be
                           "circle". Why? Who knows.
        """
        return self._set_adjust(action, prop, **kwargs)

    def _set_adjust(self, action, prop, **kwargs):
        return "set_adjust", [action, prop], kwargs

    @_command
    def set_hsv(self, hue, saturation, value=None, light_type=LightType.Main, **kwargs):
        """
        Set the bulb's HSV value.

        :param int hue:        The hue to set (0-359).
        :param int saturation: The saturation to set (0-100).
        :param int value:      The value to set (0-100). If omitted, the bulb's
                               brightness will remain the same as before the
                               change.
        :param yeelight.LightType light_type: Light type to control.
        """
        self.ensure_on()
        return self._set_hsv(hue, saturation, value, light_type, **kwargs)

    def _set_hsv(self, hue, saturation, value, light_type, **kwargs):
        # We fake this using flow so we can add the `value` parameter.
        hue = _clamp(hue, 0, 359)
        saturation = _clamp(saturation, 0, 100)

        if value is None:
            # If no value was passed, use ``set_hsv`` to preserve luminance.
            return "set_hsv", [hue, saturation], dict(kwargs, light_type=light_type)
        else:
            # Otherwise, use flow.
            value = _clamp(value, 0, 100)

            if kwargs.get("effect", self.effect) == "sudden":
                duration = 50
            else:
                duration = kwargs.get("duration", self.duration)

            hue = _clamp(hue, 0, 359) / 359.0
            saturation = _clamp(saturation, 0, 100) / 100.0
            rgb = rgb_to_yeelight(
                *[
                    int(round(col * 255))
                    for col in colorsys.hsv_to_rgb(hue, saturation, 1)
                ]
            )
            return (
                "start_cf",
                [1, 1, "%s, 1, %s, %s" % (duration, rgb, value)],
                dict(kwargs, light_type=light_type),
            )

    @_command
    def set_brightness(self, brightness, light_type=LightType.Main, **kwargs):
        """
        Set the bulb's brightness.

        :param int brightness: The brightness value to set (1-100).
        :param yeelight.LightType light_type: Light type to control.
        """
        self.ensure_on()
        return self._set_brightness(brightness, light_type=light_type, **kwargs)

    def _set_brightness(self, brightness, light_type=LightType.Main, **kwargs):
        brightness = _clamp(brightness, 1, 100)
        return "set_bright", [brightness], dict(kwargs, light_type=light_type)

    @_command
    def turn_on(self, light_type=LightType.Main, **kwargs):
        """
        Turn the bulb on.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._turn_on(light_type=light_type, **kwargs)

    def _turn_on(self, light_type=LightType.Main, **kwargs):
        return "set_power", ["on"], dict(kwargs, light_type=light_type)

    @_command
    def turn_off(self, light_type=LightType.Main, **kwargs):
        """
        Turn the bulb off.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._turn_off(light_type=light_type, **kwargs)

    def _turn_off(self, light_type=LightType.Main, **kwargs):
        return "set_power", ["off"], dict(kwargs, light_type=light_type)

    @_command
    def toggle(self, light_type=LightType.Main, **kwargs):
        """
        Toggle the bulb on or off.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._toggle(light_type=light_type, **kwargs)

    def _toggle(self, light_type=LightType.Main, **kwargs):
        return "toggle", [], dict(kwargs, light_type=light_type)

    @_command
    def dev_toggle(self, **kwargs):
        """Toggle the main light and the ambient on or off."""
        return self._dev_toggle(**kwargs)

    def _dev_toggle(self, **kwargs):
        return "dev_toggle", [], kwargs

    @_command
    def set_default(self, light_type=LightType.Main, **kwargs):
        """
        Set the bulb's current state as the default, which is what the bulb will be set to on power on.

        If you get a "general error" setting this, yet the bulb reports as supporting `set_default` during
        discovery, disable "auto save settings" in the YeeLight app.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._set_default(light_type=light_type, **kwargs)

    def _set_default(self, light_type=LightType.Main, **kwargs):
        return "set_default", [], dict(kwargs, light_type=light_type)

    @_command
    def set_name(self, name, **kwargs):
        """
        Set the bulb's name.

        :param str name: The string you want to set as the bulb's name.
        """
        return self._set_name(name, **kwargs)

    def _set_name(self, name, **kwargs):
        return "set_name", [name], kwargs

    @_command
    def start_flow(self, flow, light_type=LightType.Main, **kwargs):
        """
        Start a flow.

        :param yeelight.Flow flow: The Flow instance to start.
        """
        self.ensure_on()
        return self._start_start_flow(flow, light_type=light_type, **kwargs)

    def _start_start_flow(self, flow, light_type=LightType.Main, **kwargs):
        if not isinstance(flow, Flow):
            raise ValueError("Argument is not a Flow instance.")

        return (
            "start_cf",
            flow.as_start_flow_params,
            dict(kwargs, light_type=light_type),
        )

    @_command
    def stop_flow(self, light_type=LightType.Main, **kwargs):
        """
        Stop a flow.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._stop_flow(light_type=light_type, **kwargs)

    def _stop_flow(self, light_type=LightType.Main, **kwargs):
        return "stop_cf", [], dict(kwargs, light_type=light_type)

    @_command
    def set_scene(self, scene_class, *args, light_type=LightType.Main, **kwargs):
        """
        Set the light directly to the specified state.

        If the light is off, it will first be turned on.

        :param yeelight.SceneClass scene_class: The YeeLight scene class to use.

        * `COLOR` changes the light to the specified RGB color and brightness.

            Arguments:
            * **red** (*int*)         – The red value to set (0-255).
            * **green** (*int*)       – The green value to set (0-255).
            * **blue** (*int*)        – The blue value to set (0-255).
            * **brightness** (*int*)  – The brightness value to set (1-100).

        * `HSV` changes the light to the specified HSV color and brightness.

            Arguments:
            * **hue** (*int*)         – The hue to set (0-359).
            * **saturation** (*int*)  – The saturation to set (0-100).
            * **brightness** (*int*)  – The brightness value to set (1-100).

        * `CT` changes the light to the specified color temperature.

            Arguments:
            * **degrees** (*int*)     – The degrees to set the color temperature to (min/max are specified by the
            model's capabilities, or 1700-6500).
            * **brightness** (*int*)  – The brightness value to set (1-100).

        * `CF` starts a color flow.

            Arguments:
            * **flow** (`yeelight.Flow`)  – The Flow instance to start.

        * `AUTO_DELAY_OFF` turns the light on to the specified brightness and sets a timer to turn it back off after the
          given number of minutes.

            Arguments:
            * **brightness** (*int*)     – The brightness value to set (1-100).
            * **minutes** (*int*)        – The minutes to wait before automatically turning the light off.

        :param yeelight.LightType light_type: Light type to control.
        """
        return self._set_scene(scene_class, *args, light_type=light_type, **kwargs)

    def _set_scene(self, scene_class, *args, light_type=LightType.Main, **kwargs):
        scene_args = [scene_class.name.lower()]
        if scene_class == SceneClass.COLOR:
            scene_args += [rgb_to_yeelight(*args[:3]), args[3]]
        elif scene_class == SceneClass.HSV:
            scene_args += args
        elif scene_class == SceneClass.CT:
            scene_args += [self._clamp_color_temp(args[0]), args[1]]
        elif scene_class == SceneClass.CF:
            scene_args += args[0].as_start_flow_params
        elif scene_class == SceneClass.AUTO_DELAY_OFF:
            scene_args += args
        else:
            raise ValueError(
                "Scene class argument is unknown. Please use one from yeelight.SceneClass."
            )

        return "set_scene", scene_args, dict(kwargs, light_type=light_type)

    def start_music(self, port=0, ip=None):
        """
        Start music mode.

        Music mode essentially upgrades the existing connection to a reverse one
        (the bulb connects to the library), removing all limits and allowing you
        to send commands without being rate-limited.

        Starting music mode will start a new listening socket, tell the bulb to
        connect to that, and then close the old connection. If the bulb cannot
        connect to the host machine for any reason, bad things will happen (such
        as library freezes).

        :param int port: The port to listen on. If none is specified, a random
                         port will be chosen.

        :param str ip: The IP address of the host this library is running on.
                       Will be discovered automatically if not provided.
        """
        if self._music_mode:
            raise AssertionError("Already in music mode, please stop music mode first.")

        # Force populating the cache in case we are being called directly
        # without ever fetching properties beforehand
        self.get_properties()

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # Reuse sockets so we don't hit "address already in use" errors.
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(("", port))
        host, port = s.getsockname()
        s.listen(3)

        local_ip = ip if ip else self._socket.getsockname()[0]
        self.send_command("set_music", [1, local_ip, port])
        s.settimeout(5)
        conn, _ = s.accept()
        s.close()  # Close the listening socket.
        self.__socket.close()
        self.__socket = conn
        self._music_mode = True

        return "ok"

    @_command
    def stop_music(self, **kwargs):
        """
        Stop music mode.

        Stopping music mode will close the previous connection. Calling
        ``stop_music`` more than once, or while not in music mode, is safe.
        """
        if self.__socket:
            self.__socket.close()
            self.__socket = None
        self._music_mode = False
        return "set_music", [0], kwargs

    @_command
    def cron_add(self, event_type, value, **kwargs):
        """
        Add an event to cron.

        Example::

        >>> bulb.cron_add(CronType.off, 10)

        :param yeelight.CronType event_type: The type of event. Currently,
                                                   only ``CronType.off``.
        """
        return self._cron_add(event_type, value, **kwargs)

    def _cron_add(self, event_type, value, **kwargs):
        return "cron_add", [event_type.value, value], kwargs

    @_command
    def cron_get(self, event_type, **kwargs):
        """
        Retrieve an event from cron.

        :param yeelight.CronType event_type: The type of event. Currently,
                                                   only ``CronType.off``.
        """
        return self._cron_get(event_type, **kwargs)

    def _cron_get(self, event_type, **kwargs):
        return "cron_get", [event_type.value], kwargs

    @_command
    def cron_del(self, event_type, **kwargs):
        """
        Remove an event from cron.

        :param yeelight.CronType event_type: The type of event. Currently,
                                                   only ``CronType.off``.
        """
        return self._cron_del(event_type, **kwargs)

    def _cron_del(self, event_type, **kwargs):
        return "cron_del", [event_type.value], kwargs

    def __repr__(self):
        return "Bulb<{ip}:{port}, type={type}>".format(
            ip=self._ip, port=self._port, type=self.bulb_type
        )

    def set_power_mode(self, mode):
        """
        Set the light power mode.

        If the light is off it will be turned on.

        :param yeelight.PowerMode mode: The mode to switch to.
        """
        return self.turn_on(power_mode=mode)

    def get_model_specs(self, **kwargs):
        """Return the specifications (e.g. color temperature min/max) of the bulb."""
        if self.model is not None and self.model in _MODEL_SPECS:
            return _MODEL_SPECS[self.model]

        _LOGGER.debug("Model unknown (%s). Providing a fallback", self.model)
        if self.bulb_type is BulbType.White:
            return _MODEL_SPECS["mono"]

        if self.bulb_type is BulbType.WhiteTemp:
            return _MODEL_SPECS["ceiling1"]

        if self.bulb_type is BulbType.WhiteTempMood:
            return _MODEL_SPECS["ceiling4"]

        # BulbType.Color and BulbType.Unknown
        return _MODEL_SPECS["color"]

    def _clamp_color_temp(self, degrees):
        """
        Clamp color temp to correct range.

        :param int degrees: The degrees to set the color temperature to specified by model or defaults
        (1700-6500).
        """
        if self.model:
            color_specs = self.get_model_specs()["color_temp"]
            return _clamp(degrees, color_specs["min"], color_specs["max"])

        return _clamp(degrees, 1700, 6500)



import socket

import ifaddr


def get_ip_address(ifname):
    """
    Return the first IPv4 address of the requested interface.

    :param string interface: The interface to get the IPv4 address of.

    :returns: The interface's IPv4 address.

    """
    for adapter in ifaddr.get_adapters():
        if adapter.name != ifname:
            continue
        for ip in adapter.ips:
            if not isinstance(ip.ip, tuple):  # not IPv6
                return ip.ip
    return None


def send_discovery_packet(timeout=2, interface=False, ip_address="239.255.255.250"):
    """
    Send SSDP discovery packet.

    :param int timeout: How many seconds to wait for replies. Discovery will
                        always take exactly this long to run, as it can't know
                        when all the bulbs have finished responding.
    :param string interface: The interface that should be used for multicast packets.
                             Note: it *has* to have a valid IPv4 address. IPv6-only
                             interfaces are not supported (at the moment).
                             The default one will be used if this is not specified.
    :param string ip_address: IP address to send ssdp discovery packet to. If provided, it will be send to specified
                              device. Otherwise it will be sent to the multicast address.

    :return: Socket used to send packet.

    """
    msg = "\r\n".join(
        [
            "M-SEARCH * HTTP/1.1",
            "HOST: " + ip_address + ":1982",
            'MAN: "ssdp:discover"',
            "ST: wifi_bulb",
        ]
    )

    # Set up the UDP socket.
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 32)
    if interface:
        s.setsockopt(
            socket.IPPROTO_IP,
            socket.IP_MULTICAST_IF,
            socket.inet_aton(get_ip_address(interface)),
        )
    s.settimeout(timeout)
    s.sendto(msg.encode(), (ip_address, 1982))

    return s


def parse_capabilities(data):
    """
    Parse SSDP discovery capabilities to a dictionary.

    :param string data: Original data from SSDP discovery from the bulb.

    Example:
    'HTTP/1.1 200 OK
    Cache-Control: max-age=3600
    Date:
    Ext:
    Location: yeelight://10.0.7.184:55443
    Server: POSIX UPnP/1.0 YGLC/1
    id: 0x00000000037073d2
    model: color
    fw_ver: 76
    ...'

    :return: Parsed response as dict.

    Example:
    {
        'Location': 'yeelight://10.0.7.184:55443',
        'Server': 'POSIX UPnP/1.0 YGLC/1',
        'id': '0x00000000037073d2',
        'model': 'color',
        'fw_ver': '76',
        ...
    }
    """
    return dict(
        [x.strip("\r").split(": ") for x in data.decode().split("\n") if ":" in x]
    )


def filter_lower_case_keys(dict):
    """
    Filter dict to include only lower case keys.

    Used to skip HTTP response fields.

    :param dict: Dict with all capabilities parsed from the SSDP discovery.

    :return: Dict with lower case keys only.
    """
    return {key: value for key, value in dict.items() if key.islower()}



"""Built-in flows which the vendor ships in their app and pre-made transitions, for your strobing pleasure."""
from .flow import Action
from .flow import Flow
from .flow import RGBTransition
from .flow import TemperatureTransition
from yeelight import transitions


def disco(bpm=120):
    """
    Color changes to the beat.

    :param int bpm: The beats per minute to pulse to.

    :returns: An infinite Flow consisting of 8 transitions.
    :rtype: Flow
    """
    return Flow(count=0, action=Action.recover, transitions=transitions.disco(bpm))


def temp():
    """
    Slowly-changing color temperature.

    :returns: An infinite Flow consisting of 2 transitions.
    :rtype: Flow
    """
    return Flow(count=0, action=Action.recover, transitions=transitions.temp())


def strobe():
    """
    Rapid flashing on and off.

    :returns: An infinite Flow consisting of 2 transitions.
    :rtype: Flow
    """
    return Flow(count=0, action=Action.recover, transitions=transitions.strobe())


def pulse(red, green, blue, duration=250, brightness=100, count=1):
    """
    Pulse a single color once (mainly to be used for notifications).

    :param int red: The red color component to pulse (0-255).
    :param int green: The green color component to pulse (0-255).
    :param int blue: The blue color component to pulse (0-255).
    :param int duration: The duration to pulse for, in milliseconds.
    :param int brightness: The brightness to pulse at (1-100).
    :param int count: The number times to pulse.

    :returns: A Flow consisting of 2 transitions, after which the bulb returns to its previous state.
    :rtype: Flow
    """
    return Flow(
        count,
        action=Action.recover,
        transitions=transitions.pulse(red, green, blue, duration, brightness),
    )


def strobe_color(brightness=100):
    """
    Rapid flashing colors.

    :param int brightness: The brightness of the transition.

    :returns: An infinite Flow consisting of 6 transitions.
    :rtype: Flow
    """
    return Flow(
        count=0, action=Action.recover, transitions=transitions.strobe_color(brightness)
    )


def alarm(duration=250):
    """
    Red alarm; flashing bright red to dark red.

    :param int duration: The duration between hi/lo brightness,in milliseconds.

    :returns: An infinite Flow consisting of 2 transitions.
    :rtype: Flow
    """
    return Flow(count=0, action=Action.recover, transitions=transitions.alarm(duration))


def police(duration=300, brightness=100):
    """
    Color changes from red to blue, like police lights.

    :param int duration: The duration between red and blue, in milliseconds.
    :param int brightness: The brightness of the transition.

    :returns: An infinite Flow consisting of 2 transitions.
    :rtype: Flow
    """
    return Flow(
        count=0,
        action=Action.recover,
        transitions=transitions.police(duration, brightness),
    )


def police2(duration=250, brightness=100):
    """
    Color flashes red and then blue, like urgent police lights.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.

    :returns: An infinite Flow consisting of 8 transitions.
    :rtype: Flow
    """
    return Flow(
        count=0,
        action=Action.recover,
        transitions=transitions.police2(duration, brightness),
    )


def lsd(duration=3000, brightness=100):
    """
    Gradual changes to a pleasing, trippy palette.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.

    :returns: An infinite Flow consisting of 5 transitions.
    :rtype: Flow
    """
    return Flow(
        count=0,
        action=Action.recover,
        transitions=transitions.lsd(duration, brightness),
    )


def christmas(duration=250, brightness=100, sleep=3000):
    """
    Color changes from red to green, like christmas lights.

    :param int duration: The duration between red and green, in milliseconds.
    :param int brightness: The brightness of the transition.
    :param int sleep: The time to sleep between colors, in milliseconds.

    :returns: An infinite Flow consisting of 4 transitions.
    :rtype: Flow
    """
    return Flow(
        count=0,
        action=Action.recover,
        transitions=transitions.christmas(duration, brightness, sleep),
    )


def rgb(duration=250, brightness=100, sleep=3000):
    """
    Color changes from red to green to blue.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.
    :param int sleep: The time to sleep between colors, in milliseconds

    :returns: An infinite Flow consisting of 6 transitions.
    :rtype: Flow
    """
    return Flow(
        count=0,
        action=Action.recover,
        transitions=transitions.rgb(duration, brightness, sleep),
    )


def random_loop(duration=750, brightness=100, count=9):
    """
    Color changes between `count` randomly chosen colors.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.
    :param int count: The number of random chosen colors in transition.

    :returns: An infinite Flow consisting of up to 9 transitions.
    :rtype: Flow
    """
    return Flow(
        count=0,
        action=Action.recover,
        transitions=transitions.random_loop(duration, brightness, count),
    )


def slowdown(duration=2000, brightness=100, count=8):
    """
    Changes between `count` random chosen colors with increasing transition time.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.
    :param int count: The number of random chosen colors in transition.

    :returns: An infinite Flow consisting of up to 8 transitions.
    :rtype: Flow
    """
    return Flow(
        count=0,
        action=Action.recover,
        transitions=transitions.slowdown(duration, brightness, count),
    )


def home(duration=500, brightness=80):
    """
    Simulate daylight.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.

    :returns: An infinite Flow consisting of 1 transition.
    :rtype: Flow
    """
    transition = [
        TemperatureTransition(degrees=3200, duration=duration, brightness=brightness)
    ]
    return Flow(count=0, action=Action.recover, transitions=transition)


def night_mode(duration=500, brightness=1):
    """
    Dim the lights to a dark red, pleasant for the eyes at night.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.

    :returns: An infinite Flow consisting of 1 transition.
    :rtype: Flow
    """
    transition = [
        RGBTransition(0xFF, 0x99, 0x00, duration=duration, brightness=brightness)
    ]
    return Flow(count=0, action=Action.recover, transitions=transition)


def date_night(duration=500, brightness=50):
    """
    Dim the lights to a cozy orange.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.

    :returns: An infinite Flow consisting of 1 transition.
    :rtype: Flow
    """
    transition = [
        RGBTransition(0xFF, 0x66, 0x00, duration=duration, brightness=brightness)
    ]
    return Flow(count=0, action=Action.recover, transitions=transition)


def movie(duration=500, brightness=50):
    """
    Dim the lights to a comfy purple.

    :param int duration: The duration to fade to next color, in milliseconds.
    :param int brightness: The brightness of the transition.

    :returns: An infinite Flow consisting of 1 transition.
    :rtype: Flow
    """
    transition = [
        RGBTransition(
            red=0x14, green=0x14, blue=0x32, duration=duration, brightness=brightness
        )
    ]
    return Flow(count=0, action=Action.recover, transitions=transition)


def sunrise():
    """
    Simulate a natural and gentle sunrise in 15 minutes.

    :returns: A Flow consisting of 3 transitions, after which the bulb stays on.
    :rtype: Flow
    """
    transitions = [
        RGBTransition(red=0xFF, green=0x4D, blue=0x00, duration=50, brightness=1),
        TemperatureTransition(degrees=1700, duration=360000, brightness=10),
        TemperatureTransition(degrees=2700, duration=540000, brightness=100),
    ]
    return Flow(count=1, action=Action.stay, transitions=transitions)


def sunset():
    """
    Simulate a natural sunset and offering relaxing dimming to help you sleep in 10 minutes.

    :returns: A Flow consisting of 3 transitions, after which the bulb turns off.
    :rtype: Flow
    """
    transitions = [
        TemperatureTransition(degrees=2700, duration=50, brightness=10),
        TemperatureTransition(degrees=1700, duration=180000, brightness=5),
        RGBTransition(red=0xFF, green=0x4C, blue=0x00, duration=420000, brightness=1),
    ]
    return Flow(count=1, action=Action.off, transitions=transitions)


def romance():
    """
    Romantic lights.

    :returns: An infinite Flow consisting of 2 transitions.
    :rtype: Flow
    """
    transitions = [
        RGBTransition(red=0x59, green=0x15, blue=0x6D, duration=4000, brightness=1),
        RGBTransition(red=0x66, green=0x14, blue=0x2A, duration=4000, brightness=1),
    ]
    return Flow(count=0, action=Action.stay, transitions=transitions)


def happy_birthday():
    """
    Happy Birthday lights.

    :returns: An infinite Flow consisting of 3 transitions.
    :rtype: Flow
    """
    transitions = [
        RGBTransition(red=0xDC, green=0x50, blue=0x19, duration=1996, brightness=80),
        RGBTransition(red=0xDC, green=0x78, blue=0x1E, duration=1996, brightness=80),
        RGBTransition(red=0xAA, green=0x32, blue=0x14, duration=1996, brightness=80),
    ]
    return Flow(count=0, action=Action.stay, transitions=transitions)


def candle_flicker():
    """
    Simulate candle flicker.

    :returns: An infinite Flow consisting of 9 transitions.
    :rtype: Flow
    """
    transitions = [
        TemperatureTransition(degrees=2700, duration=800, brightness=50),
        TemperatureTransition(degrees=2700, duration=800, brightness=30),
        TemperatureTransition(degrees=2700, duration=1200, brightness=80),
        TemperatureTransition(degrees=2700, duration=800, brightness=60),
        TemperatureTransition(degrees=2700, duration=1200, brightness=90),
        TemperatureTransition(degrees=2700, duration=2400, brightness=50),
        TemperatureTransition(degrees=2700, duration=1200, brightness=80),
        TemperatureTransition(degrees=2700, duration=800, brightness=60),
        TemperatureTransition(degrees=2700, duration=400, brightness=70),
    ]
    return Flow(count=0, action=Action.recover, transitions=transitions)
